(use test srfi-1 srfi-13)
(use coops coops-introspection coops-extras)

(define-class <s1xy> () ((x 's1xy-x) (y 's1xy-y)))
(define-class <s1ab> () ((a 's1ab-a) (b 's1ab-b)))
(define-class <s2xz> () ((x 's2xz-x) (z 's2xz-z)))
(define-class <s2ac> () ((a 's2ac-a) (c 's2ac-c)))

(define-class <s1xys1ab> (<s1xy> <s1ab>)
  ((x 's1xys1ab-x) (y 's1xys1ab-y) (a 's1xys1ab-a) (b 's1xys1ab-b) (me '<s1xys1ab>)))
(define-class <s2xzs2ac> (<s2xz> <s2ac>)
  ((x 's2xzs2ac-x) (z 's2xzs2ac-z) (a 's2xzs2ac-a) (c 's2xzs2ac-c)))

(define-class <s1xys2xz> (<s1xy> <s2xz>) ())
(define-class <s2xzs1ab> (<s2xz> <s1ab>) ())

(define-generic (foo abc xyz))

(define-method (foo primary: (abc <s1xy>) (xyz <s1ab>)) 'primary-<s1xy><s1ab>)
(define-method (foo before: (abc <s1xy>) (xyz <s1ab>)) 'before-<s1xy><s1ab>)
(define-method (foo after: (abc <s1xy>) (xyz <s1ab>)) 'after-<s1xy><s1ab>)
(define-method (foo around: (abc <s1xy>) (xyz <s1ab>)) 'around-<s1xy><s1ab>)

(define-method (foo primary: (abc <s2xz>) (xyz <s2ac>)) 'primary-<s2xz><s2ac>)
(define-method (foo before: (abc <s2xz>) (xyz <s2ac>)) 'before-<s2xz><s2ac>)
(define-method (foo after: (abc <s2xz>) (xyz <s2ac>)) 'after-<s2xz><s2ac>)
(define-method (foo around: (abc <s2xz>) (xyz <s2ac>)) 'around-<s2xz><s2ac>)

(test-assert (class? <standard-class>))

(test-assert (class? <standard-object>))

(test (list <s2xz> <standard-object> <s1ab>) (class-precedence-list <s2xzs1ab>))
(test '(x z a b) (class-slots <s2xzs1ab>))

(test-assert (class? <s2xzs2ac>))
(test (list <standard-object>) (class-direct-supers <s2ac>))
(test '(a c) (class-direct-slots <s2ac>))

(test-assert (class? <s1xys1ab>))
(test (list <s1xy> <s1ab>) (class-direct-supers <s1xys1ab>))
(test '(me) (class-direct-slots <s1xys1ab>))

(test 'foo (generic-name foo))
(test '(abc xyz) (generic-specialized-arguments foo))
(let ((primaries (generic-methods foo)))
  (for-each
    (lambda (x)
      (test-assert (method? x))
      (let ((specials (method-specializers x)))
        (test-assert (list? specials))
        (test-assert (every class? specials)) )
      (test-assert (procedure? (method-procedure x))) )
    primaries) )

(unless (zero? (test-failure-count)) (exit 1))

;instance? x
;instance-of?
;primitive-instance?

;slot@

;make/copy

;describe-object

;---
#|
;; Named (has a name) "concept"

(define-generic (name obj))
(define-class <named> () (
  (namsym #:reader name) ) )

;; Moves foreward thru a set of values "concept"

(define-generic (step-function obj))
(define-class <stepper> () (
  (nxtval #:reader step-function) ) )
(define-generic (next-value obj))
(define-method (next-value (obj <stepper>)) ((step-function obj)))

;; Parameterized extension "concept"

(define-generic (parameters obj))
(define-generic (basis obj))
(define-class <parameterized> () (
  (parms #:reader parameters)
  (src #:reader basis) ) )

;; Parameterized generative set of random values "concept"

(define-class <random-distribution> (<named> <parameterized> <stepper>) (
  temp ) )

;; Create an instance of <random-distribution> where the arguments are
;; the same as the documented procedural distribution API.
;;
;; SRFI 27 API: ({some distribution constructor} arg...)
;;      OO API: (make-random-distribution {some distribution constructor} arg...)

(define-syntax make-random-distribution
  (syntax-rules ()
    ((_ ?ctor ?arg0 ...)
      (make <random-distribution> 'temp (?ctor ?arg0 ...)) ) ) )

(define-method (initialize-instance (obj <random-distribution>))
  ; The 'ctor' should be a globally defined procedure compiled
  ; with procedure-information. So if following nomenclature then the last
  ; procedure name element will be the kind of distribution.
  (let* ((temp (slot-value obj 'temp))
         (ctor (car temp))
         (procinfo (procedure-information ctor))
         (name (and procinfo (pair? procinfo) (symbol->string (car procinfo))))
         (name (and name
                    (and-let* ((kndpos (string-index-right name #\-)))
                      (substring/shared name (fx+ kndpos 1)) ) ) )
         (dstr-vals (receive (apply ctor (cdr temp))))
         (parms (and (fx<= 2 (length dstr-vals)) (receive ((second dstr-vals))))) )
    (set! (slot-value obj 'temp) #f) ;"free" the "any" slot
    (set! (slot-value obj 'namsym) (string->symbol name))
    (set! (slot-value obj 'nxtval) (first dstr-vals))
    (set! (slot-value obj 'parms) (and parms (drop-right parms 1)))
|#
